@@ -47,6 +47,18 @@ static FILE *outfile;
 static int Graph = 0;
 int File;
 
+/*------------------------------------------------------*/
+/* Structure for stacking nested `if[n]def in verilog	*/
+/*------------------------------------------------------*/
+
+struct ifstack {
+    int invert;
+    struct property *kl;
+    struct ifstack *next;
+};
+
+struct ifstack *condstack = NULL;
+
 extern char *SetExtension(char *buffer, char *path, char *extension)
 /* add 'extension' to 'path' (overwriting previous extension, if any),
    write it into buffer (if buffer is null, malloc a buffer).
@@ -191,6 +203,8 @@ struct filestack {
 
 static struct filestack *OpenFiles = NULL;
 
+struct hashdict *definitions = (struct hashdict *)NULL;
+
 #define WHITESPACE_DELIMITER " \t\n\r"
 
 /*----------------------------------------------------------------------*/
@@ -251,7 +265,6 @@ void TrimQuoted(char *line)
 	    }
 	}
     }
-
 }
 
 /*----------------------------------------------------------------------*/
@@ -265,39 +278,202 @@ void TrimQuoted(char *line)
 
 int GetNextLineNoNewline(char *delimiter)
 {
-  char *newbuf;
-  int testc;
+    char *newbuf;
+    int testc;
+    int nested = 0;
 
-  if (feof(infile)) return -1;
+    if (feof(infile)) return -1;
 
-  // This is more reliable than feof() ...
-  testc = getc(infile);
-  if (testc == -1) return -1;
-  ungetc(testc, infile); 
+    while (1) {	    /* May loop indefinitely in an `if[n]def conditional */
 
-  if (linesize == 0) {
-	/* Allocate memory for line */
-	linesize = 500;
-	line = (char *)MALLOC(linesize);
-	linetok = (char *)MALLOC(linesize);
-  }
-  fgets(line, linesize, infile);
-  while (strlen(line) == linesize - 1) {
-       newbuf = (char *)MALLOC(linesize + 500);
-       strcpy(newbuf, line);
-       FREE(line);
-       line = newbuf;
-       fgets(line + linesize - 1, 501, infile);
-       linesize += 500;
-       FREE(linetok);
-       linetok = (char *)MALLOC(linesize);
-  }
-  linenum++;
-  strcpy(linetok, line);
-  TrimQuoted(linetok);
+	// This is more reliable than feof() ...
+	testc = getc(infile);
+	if (testc == -1) return -1;
+	ungetc(testc, infile); 
 
-  nexttok = strdtok(linetok, WHITESPACE_DELIMITER, delimiter);
-  return 0;
+	if (linesize == 0) {
+	    /* Allocate memory for line */
+	    linesize = 500;
+	    line = (char *)MALLOC(linesize);
+	    linetok = (char *)MALLOC(linesize);
+	}
+	fgets(line, linesize, infile);
+	while (strlen(line) == linesize - 1) {
+	    newbuf = (char *)MALLOC(linesize + 500);
+	    strcpy(newbuf, line);
+	    FREE(line);
+	    line = newbuf;
+	    fgets(line + linesize - 1, 501, infile);
+	    linesize += 500;
+	    FREE(linetok);
+	    linetok = (char *)MALLOC(linesize);
+	}
+
+	/* Check for substitutions (verilog only).  Make sure linetok is    */
+	/* large enough to hold the entire line after substitutions.	    */
+
+	if (definitions != NULL) {
+	    char *s, *w, e;
+	    struct property *kl;
+	    int len, dlen, vlen, addin = 0;
+	    unsigned char found = FALSE;
+
+	    for (s = line; *s != '\0'; s++) {
+		if (*s == '`') {
+		    w = s + 1;
+		    while (isalnum(*w)) w++;
+		    e = *w;
+		    *w = '\0';
+		    kl = (struct property *)HashLookup(s + 1, definitions);
+		    if (kl != NULL) {
+			dlen = strlen(s);
+			if (kl->type == PROP_STRING) {
+			    vlen = strlen(kl->pdefault.string);
+			}
+			else vlen = 12;  /* Leave room for numeric conversion */
+			addin += vlen - dlen + 1;
+			found = TRUE;
+		    }
+		    *w = e;
+		}
+	    }
+	    if (found) {
+		len = strlen(line);
+		if (len + addin > linesize) {
+		    while (len + addin > linesize) linesize += 500;
+		    FREE(linetok);
+		    linetok = (char *)MALLOC(linesize);
+		}
+	    }
+	}
+
+	/* Make definition substitutions (verilog only) */
+
+	if (definitions != NULL) {
+	    char *s, *t, *w, e;
+	    struct property *kl;
+
+	    t = linetok;
+	    for (s = line; *s != '\0'; s++) {
+		if (*s == '`') {
+		    w = s + 1;
+		    while (isalnum(*w)) w++;
+		    e = *w;
+		    *w = '\0';
+		    kl = (struct property *)HashLookup(s + 1, definitions);
+		    if (kl != NULL) {
+			if (kl->type == PROP_STRING)
+			    strcpy(t, kl->pdefault.string);
+			else if (kl->type == PROP_INTEGER)
+			    sprintf(t, "%d", kl->pdefault.ival);
+			else if (kl->type == PROP_DOUBLE)
+			    sprintf(t, "%g", kl->pdefault.dval);
+			t += strlen(t);
+			s = w - 1;
+		    }
+		    else *t++ = *s;
+		    *w = e;
+		}
+		else *t++ = *s;
+	    }
+	    *t = '\0';
+	}
+	else
+	    strcpy(linetok, line);
+
+	TrimQuoted(linetok);
+	linenum++;
+
+	nexttok = strdtok(linetok, WHITESPACE_DELIMITER, delimiter);
+	if (nexttok == NULL) return 0;
+
+	/* Handle `ifdef, `ifndef, `elsif, `else, and `endif (verilog	*/
+	/* only, where indicated by a non-NULL "definitions")		*/
+
+	if (definitions == NULL) return 0;
+
+	/* If currently skipping through a section, handle conditionals differently */
+
+	if (condstack) {
+	    if (((condstack->invert == 0) && (condstack->kl == NULL))
+			|| ((condstack->invert == 1) && (condstack->kl != NULL))) {
+		if (match(nexttok, "`ifdef") || match(nexttok, "`ifndef")) {
+		    nested++;
+		    continue;
+		}
+		else if (nested > 0) {
+		    if (match(nexttok, "`endif")) nested--;
+		    continue;
+		}
+		else if (nexttok[0] != '`') continue; 
+	    }
+	}
+
+	/* Handle conditionals (that is not being skipped over) */
+
+	if (match(nexttok, "`endif")) {
+	    if (condstack == NULL) {
+		fprintf(stderr, "Error:  `endif without corresponding `if[n]def\n");
+	    }
+	    else {
+		struct ifstack *iftop = condstack;
+		condstack = condstack->next;
+		FREE(iftop);   
+	    }
+	}
+
+	/* Note that `if[n]def may be nested. */
+    
+	else if (match(nexttok, "`ifdef") || match(nexttok, "`ifndef") ||
+		    match(nexttok, "`elsif") || match(nexttok, "`else")) {
+
+	    /* Every `ifdef or `ifndef increases condstack by 1 */
+	    if (nexttok[1] == 'i') {
+		struct ifstack *newif = (struct ifstack *)MALLOC(sizeof(struct ifstack));
+		newif->next = condstack;
+		condstack = newif;
+	    }
+	    if (condstack == NULL) {
+		fprintf(stderr, "Error:  %s without `if[n]def\n", nexttok);
+		break;
+	    }
+	    else {
+		if (match(nexttok, "`else")) {
+		    /* Invert the sense of the if[n]def scope */
+		    condstack->invert = (condstack->invert == 1) ? 0 : 1;
+		}
+		else if (match(nexttok, "`elsif")) {
+		    nexttok = strdtok(NULL, WHITESPACE_DELIMITER, delimiter);
+		    if (nexttok == NULL) {
+			fprintf(stderr, "Error:  `elsif with no conditional.\n");
+			return 0;
+		    }
+		    /* Keep the same scope but redefine the parameter */
+		    condstack->invert = 0;
+		    condstack->kl = (struct property *)HashLookup(nexttok, definitions);
+		}
+		else {
+		    condstack->invert = (nexttok[3] == 'n') ? 1 : 0;
+		    nexttok = strdtok(NULL, WHITESPACE_DELIMITER, delimiter);
+		    if (nexttok == NULL) {
+			fprintf(stderr, "Error:  %s with no conditional.\n", nexttok);
+			return 0;
+		    }
+		    condstack->kl = (struct property *)HashLookup(nexttok, definitions);
+		}
+	    }
+	}
+	else if (condstack) {
+	    if (((condstack->invert == 0) && (condstack->kl == NULL))
+			|| ((condstack->invert == 1) && (condstack->kl != NULL)))
+		continue;
+	    else
+		break;
+	}
+	else
+	    break;
+    }
+    return 0;
 }
 
 /*----------------------------------------------------------------------*/
@@ -471,6 +647,33 @@ char *strdtok(char *pstring, char *delim1, char *delim2)
 
     /* "stoken" is now set.  Now find the end of the current token */
 
+    s = stoken;
+
+    /* Special verilog rule:  If a name begins with '\', then all characters    */
+    /* are a valid part of the name until a space character is reached.   The   */
+    /* space character becomes part of the verilog name.  The remainder of the  */
+    /* name is parsed according to the rules of "delim2".                       */
+
+    /* Special verilog rule exception:  To handle the problems caused by	*/
+    /* translating verilog backslash-escaped names into other netlist formats	*/
+    /* like SPICE where the convention is strictly prohibited, I have used the	*/
+    /* convention in qflow scripts to replace the trailing space with another	*/
+    /* backslash, so the name is effectively delimited by a pair of back-	*/
+    /* slashes.  Therefore check for a space or another backslash, whichever	*/
+    /* comes first.  That will satisfy both methods.  Technically this routine	*/
+    /* should know whether it is parsing SPICE or verilog and handle the syntax	*/
+    /* accordingly (needs to be done).						*/
+
+    if (*s == '\\') {
+        while (*s != '\0') {
+	    if ((*s == ' ') || (*s == '\\')) {
+		s++;
+		break;
+	    }
+	    s++;
+	}
+    }
+
     /* Check string from position stoken.  If a character in "delim2" is found,	*/
     /* save the character in "lastdelim", null the byte at that position, and	*/
     /* return the token.  If a character in "delim1" is found, do the same but	*/
@@ -479,7 +682,7 @@ char *strdtok(char *pstring, char *delim1, char *delim2)
     /* as for "delim2" above.  If not, then set "lastdelim" to a null byte and	*/
     /* return the token.							*/	
 
-    for (s = stoken; *s; s++) {
+    for (; *s; s++) {
 	twofer = (delim2 && (*delim2 == 'X')) ? TRUE : FALSE;
 	for (s2 = ((twofer == TRUE) ? delim2 + 1 : delim2); s2 && *s2; s2++) {
 	    if (*s2 == 'X') {
