@@ -59,8 +59,9 @@ the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 // See netfile.c for explanation of delimiters.  'X'
 // separates single-character delimiters from two-character delimiters.
-#define VLOG_DELIMITERS "X///**/#((**)X,::(){}[]="
+#define VLOG_DELIMITERS "X///**/#((**)X,;:(){}[]="
 #define VLOG_PIN_NAME_DELIMITERS "X///**/(**)X()"
+#define VLOG_PIN_CHECK_DELIMITERS "X///**/(**)X,;(){}"
 
 // Global storage for verilog parameters
 struct hashdict verilogparams;
@@ -120,19 +121,33 @@ int GetBusTok(struct bus *wb)
     if (match(nexttok, "[")) {
 	SkipTokComments(VLOG_DELIMITERS);
 
-	// Check for parameter names and substitute values if found.
-	if (nexttok[0] == '`') {
-	    kl = (struct property *)HashLookup(nexttok + 1, &verilogdefs);
+	result = sscanf(nexttok, "%d", &start);
+	if (result != 1) {
+	    char *aptr = NULL;
+	    char addin;
+
+	    // Check for "+/-(n)" at end of a parameter name
+	    aptr = strrchr(nexttok, '+');
+	    if (aptr == NULL) aptr = strrchr(nexttok, '-');
+	    if (aptr != NULL) {
+		addin = *aptr;
+		*aptr = '\0';
+	    }
+
+	    // Is name in the parameter list?
+	    kl = (struct property *)HashLookup(nexttok, &verilogparams);
 	    if (kl == NULL) {
-		Printf("Unknown definition %s found in array notation.\n", nexttok);
+		Printf("Array value %s is not a number or a parameter.\n",
+				nexttok);
+		return 1;
 	    }
 	    else {
 		if (kl->type == PROP_STRING) {
 		    result = sscanf(kl->pdefault.string, "%d", &start);
 		    if (result != 1) {
-		        Printf("Cannot parse first digit from parameter %s value %s\n",
-				nexttok, kl->pdefault.string);
-		        return 1;
+		        Printf("Parameter %s has value %s that cannot be parsed"
+				" as an integer.\n", nexttok, kl->pdefault.string);
+			return 1;
 		    }
 		}
 		else if (kl->type == PROP_INTEGER) {
@@ -141,54 +156,25 @@ int GetBusTok(struct bus *wb)
 		else if (kl->type == PROP_DOUBLE) {
 		    start = (int)kl->pdefault.dval;
 		    if ((double)start != kl->pdefault.dval) {
-		        Printf("Cannot parse first digit from parameter %s value %g\n",
-				nexttok, kl->pdefault.dval);
-		        return 1;
+		        Printf("Parameter %s has value %g that cannot be parsed"
+				" as an integer.\n", nexttok, kl->pdefault.dval);
+			return 1;
 		    }
 		}
 		else {
-		    Printf("Parameter %s has unknown type; don't know how to parse.\n",
-				nexttok);
+		    Printf("Parameter %s has unknown type; don't know how"
+				" to parse.\n", nexttok);
 		    return 1;
 		}
 	    }
-	}
-	else {
-	    result = sscanf(nexttok, "%d", &start);
-	    if (result != 1) {
-		// Is name in the parameter list?
-	        kl = (struct property *)HashLookup(nexttok, &verilogparams);
-		if (kl == NULL) {
-		    Printf("Array value %s is not a number or a parameter.\n",
-				nexttok);
+	    if (aptr != NULL) {
+		int addval;
+		*aptr = addin;
+		if (sscanf(aptr + 1, "%d", &addval) != 1) {
+		    Printf("Unable to parse parameter increment '%s'\n", aptr);
 		    return 1;
 		}
-		else {
-		    if (kl->type == PROP_STRING) {
-		        result = sscanf(kl->pdefault.string, "%d", &start);
-		        if (result != 1) {
-		            Printf("Parameter %s has value %s that cannot be parsed"
-				" as an integer.\n", nexttok, kl->pdefault.string);
-			    return 1;
-		        }
-		    }
-		    else if (kl->type == PROP_INTEGER) {
-			start = kl->pdefault.ival;
-		    }
-		    else if (kl->type == PROP_DOUBLE) {
-		        start = (int)kl->pdefault.dval;
-		        if ((double)start != kl->pdefault.dval) {
-		            Printf("Parameter %s has value %g that cannot be parsed"
-				" as an integer.\n", nexttok, kl->pdefault.dval);
-			    return 1;
-			}
-		    }
-		    else {
-		        Printf("Parameter %s has unknown type; don't know how"
-				" to parse.\n", nexttok);
-		        return 1;
-		    }
-		}
+		start += (addin == '+') ? addval : -addval;
 	    }
 	}
 	SkipTokComments(VLOG_DELIMITERS);
@@ -203,18 +189,33 @@ int GetBusTok(struct bus *wb)
 	else {
 	    SkipTokComments(VLOG_DELIMITERS);
 
-	    // Check for parameter names and substitute values if found.
-	    if (nexttok[0] == '`') {
-		kl = (struct property *)HashLookup(nexttok + 1, &verilogdefs);
+	    result = sscanf(nexttok, "%d", &end);
+	    if (result != 1) {
+		char *aptr = NULL;
+		char addin;
+
+		// Check for "+/-(n)" at end of a parameter name
+		aptr = strrchr(nexttok, '+');
+		if (aptr == NULL) aptr = strrchr(nexttok, '-');
+		if (aptr != NULL) {
+		    addin = *aptr;
+		    *aptr = '\0';
+		}
+
+		// Is name in the parameter list?
+	        kl = (struct property *)HashLookup(nexttok, &verilogparams);
 		if (kl == NULL) {
-		    Printf("Unknown definition %s found in array notation.\n", nexttok);
+		    Printf("Array value %s is not a number or a parameter.\n",
+					nexttok);
+		    return 1;
 		}
 		else {
 		    if (kl->type == PROP_STRING) {
 			result = sscanf(kl->pdefault.string, "%d", &end);
-		        if (result != 1) {
-			    Printf("Cannot parse second digit from parameter "
-					"%s value %s\n", nexttok, kl->pdefault.string);
+			if (result != 1) {
+		            Printf("Parameter %s has value %s that cannot be parsed"
+					" as an integer.\n", nexttok,
+					kl->pdefault.string);
 			    return 1;
 			}
 		    }
@@ -231,48 +232,18 @@ int GetBusTok(struct bus *wb)
 		    }
 		    else {
 		        Printf("Parameter %s has unknown type; don't know how"
-				" to parse.\n", nexttok);
+					" to parse.\n", nexttok);
 		        return 1;
 		    }
 		}
-	    }
-	    else {
-		result = sscanf(nexttok, "%d", &end);
-		if (result != 1) {
-		    // Is name in the parameter list?
-	            kl = (struct property *)HashLookup(nexttok, &verilogparams);
-		    if (kl == NULL) {
-			Printf("Array value %s is not a number or a parameter.\n",
-					nexttok);
+		if (aptr != NULL) {
+		    int addval;
+		    *aptr = addin;
+		    if (sscanf(aptr + 1, "%d", &addval) != 1) {
+			Printf("Unable to parse parameter increment '%s'\n", aptr);
 			return 1;
 		    }
-		    else {
-			if (kl->type == PROP_STRING) {
-			    result = sscanf(kl->pdefault.string, "%d", &end);
-			    if (result != 1) {
-		                Printf("Parameter %s has value %s that cannot be parsed"
-					" as an integer.\n", nexttok,
-					kl->pdefault.string);
-			        return 1;
-			    }
-			}
-			else if (kl->type == PROP_INTEGER) {
-			    end = kl->pdefault.ival;
-			}
-		        else if (kl->type == PROP_DOUBLE) {
-		            end = (int)kl->pdefault.dval;
-		            if ((double)end != kl->pdefault.dval) {
-			        Printf("Cannot parse second digit from parameter "
-					"%s value %g\n", nexttok, kl->pdefault.dval);
-			        return 1;
-			    }
-			}
-			else {
-		            Printf("Parameter %s has unknown type; don't know how"
-					" to parse.\n", nexttok);
-		            return 1;
-			}
-		    }
+		    end += (addin == '+') ? addval : -addval;
 		}
 	    }
 	}
@@ -312,7 +283,7 @@ int GetBusTok(struct bus *wb)
 
 int GetBus(char *astr, struct bus *wb)
 {
-    char *colonptr, *brackstart, *brackend;
+    char *colonptr, *brackstart, *brackend, *sigend, sdelim;
     int result, start, end;
 
     if (wb == NULL) return 0;
@@ -320,6 +291,40 @@ int GetBus(char *astr, struct bus *wb)
         wb->start = -1;
         wb->end = -1;
     }
+
+    /* Check for wire bundles.  If there are bundles, process each  */
+    /* section separately and concatenate the sizes.		    */
+    /* To be done:  Handle nested bundles, including N-times concatenation */
+
+    if (*astr == '{') {
+	struct bus wbb;
+
+	astr++;
+	wb->end = 0;
+	while((*astr != '\0') && (*astr != '}')) {
+	    sigend = strchr(astr, ',');
+	    if (sigend == NULL) sigend = strchr(astr, '}');
+	    if (sigend == NULL) {
+		Printf("Badly formed wire bundle \"%s\"\n", astr - 1);
+		return 1;
+	    }
+	    sdelim = *sigend;
+	    *sigend = '\0';
+	    if (GetBus(astr, &wbb) == 0) {
+		if (wbb.start > wbb.end)
+		    wb->start += (wbb.start - wbb.end + 1);
+		else
+		    wb->start += (wbb.end - wbb.start + 1);
+	    }
+	    else {
+		wb->start++;
+	    }
+	    *sigend = sdelim;
+	    astr = sigend + 1;
+	}
+	return 0;
+    }
+
     brackstart = strchr(astr, '[');
     if (brackstart != NULL) {
 	brackend = strchr(astr, ']');
@@ -595,7 +600,7 @@ void ReadVerilogFile(char *fname, int filenum, struct cellstack **CellStackPtr,
   instname[255] = '\0';
   in_module = (char)0;
   in_param = (char)0;
-  
+
   while (!EndParseFile()) {
 
     SkipTokComments(VLOG_DELIMITERS); /* get the next token */
@@ -620,6 +625,81 @@ void ReadVerilogFile(char *fname, int filenum, struct cellstack **CellStackPtr,
 	}
     }
 
+    /* Handle parameters by treating as a localparam or definition.	*/
+    /* Currently anything other than a constant value is not handled	*/
+    /* and so will flag a warning.					*/
+
+    else if (match(nexttok, "parameter") || match(nexttok, "localparam")) {
+	char *paramkey = NULL;
+	char *paramval = NULL;
+
+	// Pick up key = value pairs and store in current cell.  Look only
+	// at the keyword before "=".  Then set the defition as everything
+	// remaining in the line, excluding comments, until the end-of-statement
+
+	while (nexttok != NULL)
+	{
+	    struct property *kl = NULL;
+
+	    /* Parse for parameters used in expressions.  Save	*/
+	    /* parameters in the "verilogparams" hash table.	*/
+
+	    SkipTok(VLOG_DELIMITERS);
+	    if ((nexttok == NULL) || (nexttok[0] == '\0')) break;
+	    if (match(nexttok, "=")) {
+		/* Pick up remainder of statement */
+		while (nexttok != NULL) {
+		    SkipTokNoNewline("X///**/X;,");
+		    if (nexttok == NULL) break;
+		    if (match(nexttok, ";") || match(nexttok, ",")) break;
+		    if (paramval == NULL) paramval = strsave(nexttok);
+		    else {
+			char *paramlast;
+			/* Append nexttok to paramval */
+			paramlast = paramval;
+			paramval = (char *)MALLOC(strlen(paramlast) + strlen(nexttok)
+				+ 2);
+			sprintf(paramval, "%s %s", paramlast, nexttok);
+			FREE(paramlast);
+		    }
+		}
+
+		kl = NewProperty();
+		kl->key = strsave(paramkey);
+		kl->idx = 0;
+		kl->merge = MERGE_NONE;
+
+		if (ConvertStringToInteger(paramval, &ival) == 1) {
+		    kl->type = PROP_INTEGER;
+		    kl->slop.ival = 0;
+		    kl->pdefault.ival = ival;
+		}
+		else if (ConvertStringToFloat(paramval, &dval) == 1) {
+		    kl->type = PROP_DOUBLE;
+		    kl->slop.dval = 0.01;
+		    kl->pdefault.dval = dval;
+		}
+		else {
+		    kl->type = PROP_STRING;
+		    kl->slop.dval = 0.0;
+		    kl->pdefault.string = strsave(paramval);
+		}
+
+		HashPtrInstall(paramkey, kl, &verilogparams);
+		FREE(paramval);
+		paramval = NULL;
+
+		if ((nexttok == NULL) || match(nexttok, ";")) break;
+	    }
+	    else {
+		if (paramkey != NULL) FREE(paramkey);
+		paramkey = strsave(nexttok);
+	    }
+	}
+	if (paramval != NULL) FREE(paramval);
+	if (paramkey != NULL) FREE(paramkey);
+    }
+
     else if (match(nexttok, "module")) {
       InitializeHashTable(&buses, OBJHASHSIZE);
       SkipTokNoNewline(VLOG_DELIMITERS);
@@ -696,7 +776,7 @@ void ReadVerilogFile(char *fname, int filenum, struct cellstack **CellStackPtr,
       inlined_decls = (char)0;
 
       if (tp != NULL) {
-	 struct bus wb;
+	 struct bus wb, *nb;
 
 	 PushStack(tp->name, CellStackPtr);
 
@@ -760,7 +840,8 @@ void ReadVerilogFile(char *fname, int filenum, struct cellstack **CellStackPtr,
 		else {
 		    if (!match(nexttok, "input") && !match(nexttok, "output") &&
 				!match(nexttok, "inout") && !match(nexttok, "real") &&
-				!match(nexttok, "logic") && !match(nexttok, "integer")) {
+				!match(nexttok, "wire") && !match(nexttok, "logic") &&
+				!match(nexttok, "integer")) {
 			if (match(nexttok, "[")) {
 			   if (GetBusTok(&wb) != 0) {
 			      // Didn't parse as a bus, so wing it
@@ -782,6 +863,12 @@ void ReadVerilogFile(char *fname, int filenum, struct cellstack **CellStackPtr,
 				    Port(portname);
 				 }
 			      }
+			      /* Also register this port as a bus */
+			      nb = NewBus();
+			      nb->start = wb.start;
+			      nb->end = wb.end;
+			      HashPtrInstall(nexttok, nb, &buses);
+
 			      wb.start = wb.end = -1;
 			   }
 			   else {
@@ -991,151 +1078,219 @@ skip_endmodule:
       }
       HashPtrInstall(kl->key, kl, &verilogdefs);
     }
-    else if (match(nexttok, "localparam")) {
-      // Pick up key = value pairs and store in current cell
-      while (nexttok != NULL)
-      {
-	 struct property *kl = NULL;
-
-	 /* Parse for parameters used in expressions.  Save	*/
-	 /* parameters in the "verilogparams" hash table.	*/
-
-	 SkipTokNoNewline(VLOG_DELIMITERS);
-	 if ((nexttok == NULL) || (nexttok[0] == '\0')) break;
-	 if ((eqptr = strchr(nexttok, '=')) != NULL) {
-	    *eqptr = '\0';
-	    kl = NewProperty();
-	    kl->key = strsave(nexttok);
-	    kl->idx = 0;
-	    kl->merge = MERGE_NONE;
-
-	    if (ConvertStringToInteger(eqptr + 1, &ival) == 1) {
-	       kl->type = PROP_INTEGER;
-	       kl->slop.ival = 0;
-	       kl->pdefault.ival = ival;
-	    }
-	    else if (ConvertStringToFloat(eqptr + 1, &dval) == 1) {
-	       kl->type = PROP_DOUBLE;
-	       kl->slop.dval = 0.01;
-	       kl->pdefault.dval = dval;
-	    }
-	    else {
-	       kl->type = PROP_STRING;
-	       kl->slop.dval = 0.0;
-	       kl->pdefault.string = strsave(eqptr + 1);
-	    }
-	    HashPtrInstall(nexttok, kl, &verilogparams);
-	 }
-      }
-    }
-
-    /* Note:  This is just the most basic processing of conditionals,	*/
-    /* although it does handle nested conditionals.			*/
-    
-    else if (match(nexttok, "`ifdef") || match(nexttok, "`ifndef")) {
-	struct property *kl;
-	int nested = 0;
-        int invert = (nexttok[3] == 'n') ? 1 : 0;
-
-	SkipTokNoNewline(VLOG_DELIMITERS);
-
-	/* To be done:  Handle boolean arithmetic on conditionals */
-
-	kl = (struct property *)HashLookup(nexttok, &verilogdefs);
-	if (((invert == 0) && (kl == NULL))
-		|| ((invert == 1) && (kl != NULL))) {
-	   /* Skip to matching `endif */
-	   while (1) {
-	      SkipNewLine(VLOG_DELIMITERS);
-	      SkipTokComments(VLOG_DELIMITERS);
-	      if (EndParseFile()) break;
-	      if (match(nexttok, "`ifdef") || match(nexttok, "`ifndef")) {
-		  nested++;
-	      }
-	      else if (match(nexttok, "`endif")) {
-		  if (nested == 0)
-		     break;
-		  else
-		     nested--;
-	      }
-	   }
-	}
-    }
-
-    else if (match(nexttok, "wire") || match(nexttok, "assign")) {	/* wire = node */
-      struct bus wb, *nb;
-      char nodename[128];
-      int is_assignment = FALSE;
-      struct objlist *lhs, *rhs;
-
-      // Several allowed uses of "assign":
-      // "assign a = b" joins two nets.
-      // "assign a = {b, c, ...}" creates a bus from components.
-      // "assign" using any boolean arithmetic is not structural verilog.
+    else if (match(nexttok, "`undef")) {
+      struct property *kl = NULL;
 
       SkipTokNoNewline(VLOG_DELIMITERS);
-      if (match(nexttok, "real")) SkipTokNoNewline(VLOG_DELIMITERS);
-      while (nexttok != NULL) {
-	 if (match(nexttok, "=")) {
-	     is_assignment = TRUE;
-	 }
-	 else if (GetBusTok(&wb) == 0) {
-	     /* Handle bus notation */
-	     SkipTokNoNewline(VLOG_DELIMITERS);
-	     if (wb.start > wb.end) {
-		for (i = wb.end; i <= wb.start; i++) {
-		    sprintf(nodename, "%s[%d]", nexttok, i);
-	            if (LookupObject(nodename, CurrentCell) == NULL)
-		        Node(nodename);
-		}
-	     }
-	     else {
-		for (i = wb.start; i <= wb.end; i++) {
-		    sprintf(nodename, "%s[%d]", nexttok, i);
-	            if (LookupObject(nodename, CurrentCell) == NULL)
-		        Node(nodename);
-		}
-	     }
-	     nb = NewBus();
-	     nb->start = wb.start;
-	     nb->end = wb.end;
-	     HashPtrInstall(nexttok, nb, &buses);
-	 }
-	 else {
-	     if (is_assignment) {
-		 /* Handle assignment statements */
-		 /* To be done:  Handle where both are bus names, and	*/
-		 /* where lhs is a bus name and rhs is a list of nets	*/
-		 if ((rhs = LookupObject(nexttok, CurrentCell)) != NULL) {
-		     join(lhs->name, rhs->name);
-		 }
-		 else {
-		     Printf("Module '%s' is not structural verilog, "
-				"making black-box.\n", model);
-		     SetClass(CLASS_MODULE);
-		     goto skip_endmodule;
-		 }
-		 is_assignment = FALSE;
-	     }
-	     else if (LookupObject(nexttok, CurrentCell) == NULL)
-	         Node(nexttok);
-		 lhs = LookupObject(nexttok, CurrentCell);
-	 }
-	 do {
-	     SkipTokNoNewline(VLOG_DELIMITERS);
-	 } while (nexttok && match(nexttok, ";"));
+      if ((nexttok == NULL) || (nexttok[0] == '\0')) break;
+
+      kl = HashLookup(nexttok, &verilogdefs);
+      if (kl != NULL) {
+	  HashDelete(nexttok, &verilogdefs);
+	  if (kl->type == PROP_STRING)
+             if (kl->pdefault.string != NULL)
+	         FREE(kl->pdefault.string);
+	  FREE(kl->key);
       }
+      /* Presumably it is not an error to undefine an undefined keyword */
+    }
+    else if (match(nexttok, "real") || match(nexttok, "integer")) {
+	Printf("Ignoring '%s' in module '%s'\n", nexttok, model);
+	/* Do not skip to end of module, as these can be in the middle of   */
+	/* I/O assignments, which need to be parsed.			    */
+	while (!match(nexttok, ";")) SkipTok("X///**/X,;");
+	continue;
+    }
+    else if (match(nexttok, "wire") || match(nexttok, "assign")) {	/* wire = node */
+	struct bus wb, wb2, *nb;
+	char nodename[128], noderoot[100];
+	int is_wire = match(nexttok, "wire");
+	int j;
+	struct objlist *lhs, *rhs;
+
+	/* Get left-hand side expression.  If this is a wire statement,	*/
+	/* then define the wire.  If is_wire is false, then the wire	*/
+	/* should already be defined.					*/
+
+	if (is_wire) {
+	    SkipTokNoNewline(VLOG_DELIMITERS);
+	    if (match(nexttok, "real"))
+		SkipTokNoNewline(VLOG_DELIMITERS);
+	    else if (match(nexttok, "logic"))
+		SkipTokNoNewline(VLOG_DELIMITERS);
+
+	    if (GetBusTok(&wb) == 0) {
+		/* Handle bus notation */
+		SkipTokNoNewline(VLOG_DELIMITERS);
+		strcpy(noderoot, nexttok);
+		if (wb.start > wb.end) {
+		    for (i = wb.end; i <= wb.start; i++) {
+			sprintf(nodename, "%s[%d]", nexttok, i);
+			if (LookupObject(nodename, CurrentCell) == NULL)
+			    Node(nodename);
+			if (i == wb.start) lhs = LookupObject(nodename, CurrentCell);
+		    }
+		}
+		else {
+		    for (i = wb.start; i <= wb.end; i++) {
+			sprintf(nodename, "%s[%d]", nexttok, i);
+			if (LookupObject(nodename, CurrentCell) == NULL)
+			    Node(nodename);
+			if (i == wb.start) lhs = LookupObject(nodename, CurrentCell);
+		    }
+		}
+		nb = NewBus();
+		nb->start = wb.start;
+		nb->end = wb.end;
+		HashPtrInstall(nexttok, nb, &buses);
+	    }
+	    else {
+		if (LookupObject(nexttok, CurrentCell) == NULL) {
+		    Node(nexttok);
+		    lhs = LookupObject(nexttok, CurrentCell);
+		}
+	    }
+	    while (1) {
+		SkipTokNoNewline(VLOG_DELIMITERS);
+		if (match(nexttok, ",")) {
+		    SkipTokComments(VLOG_DELIMITERS);
+		    if (LookupObject(nexttok, CurrentCell) == NULL) {
+			Node(nexttok);
+			lhs = LookupObject(nexttok, CurrentCell);
+		    }
+		}
+		else break;
+	    }
+	}
+	else {	    /* "assign" */
+	    SkipTokComments(VLOG_PIN_CHECK_DELIMITERS);
+	    if (GetBus(nexttok, &wb) == 0) {
+		char *aptr = strchr(nexttok, '[');
+		if (aptr != NULL) {
+		    *aptr = '\0';
+		    /* Find object of first net in bus */
+		    strcpy(noderoot, nexttok);
+		    sprintf(nodename, "%s[%d]", nexttok, wb.start);
+		    lhs = LookupObject(nodename, CurrentCell);
+		    *aptr = '[';
+		}
+	    }
+	    else {
+		lhs = LookupObject(nexttok, CurrentCell);
+	    }
+	    SkipTokComments(VLOG_DELIMITERS);
+	    if (lhs && ((!nexttok) || (!match(nexttok, "=")))) {
+		fprintf(stderr, "Empty assignment for net %s\n", lhs->name);
+	    }
+	}
+
+	/* Check for assignment statement, and handle any allowed uses.	    */
+	/* Any uses other than those mentioned below will cause the entire  */
+	/* module to be treated as a black box.				    */
+
+	// Allowed uses of "assign" for netlists:
+	//    "assign a = b" joins two nets.
+	//    "assign a = {b, c, ...}" creates a bus from components.
+	//    "assign" using any boolean arithmetic is not structural verilog.
+
+	if (nexttok && match(nexttok, "=")) {
+	    char assignname[128], assignroot[100];
+
+	    i = wb.start;
+	    while (1) {
+		SkipTokComments(VLOG_PIN_CHECK_DELIMITERS);
+		if (!nexttok) break;
+
+		if (match(nexttok, "{")) {
+		    /* RHS is a bundle */
+		    continue;
+		}
+		else if (match(nexttok, "}")) {
+		    /* End of bundle */
+		    continue;
+		}
+		else if (match(nexttok, ",")) {
+		    /* Additional signals in bundle */
+		    continue;
+		}
+		else if (match(nexttok, ";")) {
+		    /* End of assignment */
+		    break;
+		}
+		else {
+		    if (GetBus(nexttok, &wb2) == 0) {
+			char *aptr = strchr(nexttok, '[');
+			j = wb2.start;
+			if (aptr != NULL) {
+			    *aptr = '\0';
+			    strcpy(assignroot, nexttok);
+			    sprintf(assignname, "%s[%d]", nexttok, j);
+			    rhs = LookupObject(assignname, CurrentCell);
+			    *aptr = '[';
+			}
+		    }
+		    else {
+			j = -1;
+			rhs = LookupObject(nexttok, CurrentCell);
+		    }
+		    if ((lhs == NULL) || (rhs == NULL)) {
+			if (rhs != NULL) {
+			    Printf("Improper assignment;  left-hand side cannot "
+					"be parsed.\n");
+			    Printf("Right-hand side is \"%s\".\n", rhs->name);
+			    break;
+			}
+			if (lhs != NULL) {
+			    Printf("Improper assignment;  right-hand side cannot "
+					"be parsed.\n");
+			    Printf("Left-hand side is \"%s\".\n", lhs->name);
+			    /* Not parsable, probably behavioral verilog? */
+			    Printf("Module '%s' is not structural verilog, "
+				    "making black-box.\n", model);
+			    SetClass(CLASS_MODULE);
+			    goto skip_endmodule;
+			}
+		    }
+		    while (1) {
+			/* Assign bits in turn from bundle in RHS to bits of LHS    */
+			/* until bits in signal are exhausted or LHS is full.	    */
+
+			if (i != -1)
+			    sprintf(nodename, "%s[%d]", noderoot, i);
+			else
+			    sprintf(nodename, lhs->name);
+			if (j != -1)
+			    sprintf(assignname, "%s[%d]", assignroot, j);
+			else
+			    sprintf(assignname, rhs->name);
+
+			join(nodename, assignname);
+
+			if (i == wb.end) break;
+			i += (wb.end > wb.start) ? 1 : -1;
+
+			if (j == wb2.end) break;
+			j += (wb2.end > wb2.start) ? 1 : -1;
+		    }
+		}
+	    }
+	}
+	while (nexttok && !match(nexttok, ";"))
+	    SkipTokComments(VLOG_DELIMITERS);
     }
     else if (match(nexttok, "endmodule")) {
       // No action---new module is started with next 'module' statement,
       // if any.
       SkipNewLine(VLOG_DELIMITERS);
+      in_module = (char)0;	    /* Should have been done already */
     }
     else if (nexttok[0] == '`') {
-      // Ignore any other directive starting with a backtick
+      // Ignore any other directive starting with a backtick (e.g., `timescale)
       SkipNewLine(VLOG_DELIMITERS);
     }
-    else if (match(nexttok, "reg") || match(nexttok, "always")) {
+    else if (match(nexttok, "reg") || match(nexttok, "always") ||
+	    match(nexttok, "specify") || match(nexttok, "initial")) {
+      Printf("Behavioral keyword '%s' found in source.\n", nexttok);
       Printf("Module '%s' is not structural verilog, making black-box.\n", model);
       // To be done:  Remove any contents (but may not be necessary)
       // Recast as module
@@ -1145,6 +1300,7 @@ skip_endmodule:
     else {	/* module instances */
       char instancename[100], modulename[100];
       int itype, arraystart, arrayend, arraymax, arraymin;
+      char ignore;
 
       instancename[99] = '\0';
       modulename[99] = '\0';
@@ -1164,9 +1320,12 @@ skip_endmodule:
 	PushStack(fname, CellStackPtr);
       }
       
+      SkipTokComments(VLOG_DELIMITERS);
+
+nextinst:
+      ignore = FALSE;
       head = NULL;
       tail = NULL;
-      SkipTokComments(VLOG_DELIMITERS);
 
       // Next token must be '#(' (parameters) or an instance name
 
@@ -1239,19 +1398,21 @@ skip_endmodule:
 	 // Read the pin list
          while (nexttok != NULL) {
 	    SkipTokComments(VLOG_DELIMITERS);
-	    // NOTE: Deal with `ifdef et al. properly.  Ignoring for now.
-	    while (nexttok[0] == '`') {
-		SkipNewLine(VLOG_DELIMITERS);
-		SkipTokComments(VLOG_DELIMITERS);
-	    }
 	    if (match(nexttok, ")")) break;
 	    else if (match(nexttok, ",")) continue;
 
 	    // We need to look for pins of the type ".name(value)"
 
 	    if (nexttok[0] != '.') {
-	        Printf("Badly formed subcircuit pin line at \"%s\"\n", nexttok);
-	        SkipNewLine(VLOG_DELIMITERS);
+	        Printf("Warning:  Ignoring subcircuit with no pin names "
+			    "at \"%s\"\n", nexttok);
+		InputParseError(stderr);
+		while (nexttok != NULL) {
+		    SkipTokComments(VLOG_DELIMITERS);
+		    if (match(nexttok, ";")) break;
+		}
+		ignore = TRUE;
+		break;
 	    }
 	    else {
 	       new_port = (struct portelement *)CALLOC(1, sizeof(struct portelement));
@@ -1261,31 +1422,61 @@ skip_endmodule:
 	           Printf("Badly formed subcircuit pin line at \"%s\"\n", nexttok);
 	           SkipNewLine(VLOG_DELIMITERS);
 	       }
-	       SkipTokComments(VLOG_PIN_NAME_DELIMITERS);
+	       SkipTokComments(VLOG_PIN_CHECK_DELIMITERS);
 	       if (match(nexttok, ")")) {
 		  char localnet[100];
 		  // Empty parens, so create a new local node
 		  savetok = (char)1;
-		  sprintf(localnet, "_noconnect_%d_", localcount++);
+		  if (arraystart != -1) {
+		     /* No-connect on an instance array must also be an array */
+		     sprintf(localnet, "_noconnect_%d_[%d:%d]", localcount++,
+				arraystart, arrayend);
+		  }
+		  else
+		     sprintf(localnet, "_noconnect_%d_", localcount++);
 		  new_port->net = strsave(localnet);
 	       }
 	       else {
-	          new_port->net = strsave(nexttok);
+		  if (!strcmp(nexttok, "{")) {
+		     char *in_line_net = (char *)MALLOC(1);
+		     char *new_in_line_net = NULL;
+		     *in_line_net = '\0';
+		     /* In-line array---read to "}" */
+		     while (nexttok) {
+			 new_in_line_net = (char *)MALLOC(strlen(in_line_net) +
+				    strlen(nexttok) + 1);
+			 /* Roundabout way to do realloc() becase there is no REALLOC() */
+			 strcpy(new_in_line_net, in_line_net);
+			 strcat(new_in_line_net, nexttok);
+			 FREE(in_line_net);
+			 in_line_net = new_in_line_net;
+			 if (!strcmp(nexttok, "}")) break;
+			 SkipTokComments(VLOG_PIN_CHECK_DELIMITERS);
+		     }
+		     if (!nexttok) {
+			 Printf("Unterminated net in pin %s\n", in_line_net);
+		     }
+		     new_port->net = in_line_net;
+		  }
+		  else
+		     new_port->net = strsave(nexttok);
+
 		  /* Read array information along with name;  will be parsed later */
-		  SkipTokComments(VLOG_DELIMITERS);
-	          if (match(nexttok, "[")) {
+		  SkipTokComments(VLOG_PIN_CHECK_DELIMITERS);
+		  if (match(nexttok, "[")) {
 		      /* Check for space between name and array identifier */
-	              SkipTokComments(VLOG_PIN_NAME_DELIMITERS);
-	              if (!match(nexttok, ")")) {
-		          char *expnet;
-		          expnet = (char *)MALLOC(strlen(new_port->net)
-					+ strlen(nexttok) + 2);
-		          sprintf(expnet, "%s[%s", new_port->net, nexttok);
-		          FREE(new_port->net);
-		          new_port->net = expnet;
+		      SkipTokComments(VLOG_PIN_NAME_DELIMITERS);
+		      if (!match(nexttok, ")")) {
+			 char *expnet;
+			 expnet = (char *)MALLOC(strlen(new_port->net)
+				    + strlen(nexttok) + 2);
+			 sprintf(expnet, "%s[%s", new_port->net, nexttok);
+			 FREE(new_port->net);
+			 new_port->net = expnet;
 		      }
 		      SkipTokComments(VLOG_DELIMITERS);
-	          }
+		  }
+
 	          if (!match(nexttok, ")")) {
 	              Printf("Badly formed subcircuit pin line at \"%s\"\n", nexttok);
 	              SkipNewLine(VLOG_DELIMITERS);
@@ -1302,10 +1493,21 @@ skip_endmodule:
       else {
          Printf("Expected to find instance pin block but got \"%s\"\n", nexttok);
       }
-      /* Instance should end with a semicolon */
+      if (ignore == TRUE) continue;	/* moving along. . . */
+
+      /* Verilog allows multiple instances of a single cell type to be chained	*/
+      /* together with commas.							*/
+
       SkipTokComments(VLOG_DELIMITERS);
-      if (!match(nexttok, ";")) {
+      if (match(nexttok, ",")) {
+	 goto nextinst;
+      }
+
+      /* Otherwise, instance must end with a semicolon */
+    
+      else if (!match(nexttok, ";")) {
 	 Printf("Expected to find end of instance but got \"%s\"\n", nexttok);
+	 InputParseError(stderr);
       }
 
       /* Check for ignored class */
@@ -1482,14 +1684,44 @@ skip_endmodule:
 		  // Otherwise, net is bit-sliced across array of instances.
 	       }
 	       else if (wb.start > wb.end) {
+		  char *bptr, *cptr, cchar, *netname;
+		  unsigned char is_bundle = 0;
+		  struct bus wbb;
+
 		  i = wb.start;
 		  j = portstart;
+
+		  netname = scan->net;
+		  if (*netname == '{') {
+		     is_bundle = 1;
+		     netname++;
+		     cptr = strchr(netname, ',');
+		     if (cptr == NULL) cptr = strchr(netname, '}');
+		     if (cptr == NULL) cptr = netname + strlen(netname) - 1;
+		     cchar = *cptr;
+		     *cptr = '\0';
+		  }
+
+		  // Remove indexed part of scan->net
+		  if (GetBus(netname, &wbb) == 0) {
+		     i = wbb.start;
+		     if ((bptr = strchr(netname, '[')) != NULL)
+			 *bptr = '\0';
+		  }
+		  else
+		     i = -1;
+
+		  if (is_bundle) *cptr = cchar;	 /* Restore bundle delimiter */
+		  
 		  while (1) {
 	             new_port = (struct portelement *)CALLOC(1,
 				sizeof(struct portelement));
 	             sprintf(vname, "%s[%d]", scan->name, j);
 	             new_port->name = strsave(vname);
-	             sprintf(vname, "%s[%d]", scan->net, i); 
+		     if (i == -1)
+			 sprintf(vname, "%s", netname); 
+		     else
+			 sprintf(vname, "%s[%d]", netname, i); 
 	             new_port->net = strsave(vname);
 
 		     if (last == NULL)
@@ -1504,8 +1736,31 @@ skip_endmodule:
 
 		     if (portstart > portend) j--;
 		     else j++;
-		     if (wb.start > wb.end) i--;
+		     if (wbb.start > wbb.end) i--;
 		     else i++;
+
+		     if (is_bundle &&
+			    ((i == -1) ||
+			    ((wbb.start > wbb.end) && (i < wbb.end)) ||
+			    ((wbb.start < wbb.end) && (i > wbb.end)))) {
+		         if (bptr) *bptr = '[';
+
+			 netname = cptr + 1;
+		         cptr = strchr(netname, ',');
+			 if (cptr == NULL) cptr = strchr(netname, '}');
+			 if (cptr == NULL) cptr = netname + strlen(netname) - 1;
+			 cchar = *cptr;
+			 *cptr = '\0';
+
+			 if (GetBus(netname, &wbb) == 0) {
+			    i = wbb.start;
+			    if ((bptr = strchr(netname, '[')) != NULL)
+				*bptr = '\0';
+			 }
+			 else i = -1;
+
+			 *cptr = cchar;	    /* Restore delimiter */
+		     }
 		  }
 		  FREE(scan);
 		  scan = last;
@@ -1600,7 +1855,7 @@ skip_endmodule:
 		       // Instance must be an array
 		       char netname[128];
 		       int slice;
-		       if (wb.start > wb.end && arraystart > arrayend)
+		       if (wb.start >= wb.end && arraystart >= arrayend)
 			   slice = wb.start - (arraystart - i);
 		       else if (wb.start < wb.end && arraystart > arrayend)
 			   slice = wb.start + (arraystart - i);
@@ -1608,6 +1863,7 @@ skip_endmodule:
 			   slice = wb.start - (arraystart + i);
 		       else // (wb.start < wb.end && arraystart < arrayend)
 			   slice = wb.start + (arraystart + i);
+			   
 		       sprintf(netname, "%s[%d]", scanroot, slice);
 	               if (LookupObject(netname, CurrentCell) == NULL) Node(netname);
 	               join(netname, obptr->name);
@@ -1696,6 +1952,7 @@ char *ReadVerilogTop(char *fname, int *fnum, int blackbox)
 
   InitializeHashTable(&verilogparams, OBJHASHSIZE);
   InitializeHashTable(&verilogdefs, OBJHASHSIZE);
+  definitions = &verilogdefs;
 
   /* Add the pre-defined key "LVS" to verilogdefs */
 
@@ -1718,6 +1975,7 @@ char *ReadVerilogTop(char *fname, int *fnum, int blackbox)
   HashKill(&verilogparams);
   RecurseHashTable(&verilogdefs, freeprop);
   HashKill(&verilogdefs);
+  definitions = (struct hashdict *)NULL;
 
   // Record the top level file.
   if (LookupCellFile(fname, filenum) == NULL) CellDef(fname, filenum);
@@ -1778,10 +2036,10 @@ void IncludeVerilog(char *fname, int parent, struct cellstack **CellStackPtr,
 	if (strchr(fname, '.') == NULL) {
            SetExtension(name, fname, VERILOG_EXTENSION);
            filenum = OpenParseFile(name, parent);
-	}
-        if (filenum < 0) {
-           Fprintf(stderr,"Error in Verilog file include: No file %s\n",name);
-           return;
+        }
+	if (filenum < 0) {
+	   fprintf(stderr,"Error in Verilog file include: No file %s\n", fname);
+	   return;
         }    
      }
   }
